# Code_py.py:
# This class generates python code from converted parse trees.

from sys import stdout
import re
import string

def ts(depth):
	# returns string used for proper indenting to level of 'depth'
	s = ""
	for i in range(0, depth, 1):
		s = s + "    "
	return s

# This is a list of FLTK classes that have the type 'Fl_Group'
# If a class is listed here, a call to end() will be generated.
fltkBuiltinGroupClasses = [ "Fl_Group",
"Fl_Browser",
"Fl_Browser_",
"Fl_Color_Chooser",
"Fl_Double_Window",
"Fl_Gl_Window",
"Fl_Glut_Window",
"Fl_Hold_Browser",
"Fl_Menu_Window",
"Fl_Multi_Browser",
"Fl_Overlay_Window",
"Fl_Pack",
"Fl_Scroll",
"Fl_Select_Browser",
"Fl_Single_Window",
"Fl_Tabs",
"Fl_Tile",
"Fl_Window"]


# This is a list of FLTK classes that have the type 'Fl_Menu_'
fltkBuiltinMenuClasses = [ "Fl_Menu_", "Fl_Choice", "Fl_Menu_Bar", 
							"Fl_Menu_Button" ]

fltkMenuComponents = ["submenu", "menuitem"]


# Setting options on windows sometimes changes the class of the object.
# For example Fl_Slider becomes Fl_Hor_Slider if the Horizontal option is 
# set.  Most widgets have a simple subclass defined for each combination of
# options eg. Fl_Hor_Slider simply calls type() in the constructor.  The
# preferred method is to change the class of the object based on the options.
# Unfortunately, not all widgets have the subclasses defined, that is sometimes
# fluid just keeps the main class and generates a call to type().  The following
# dictionary of dictionaries defines the mappings.  If a object has the 
# 'type' option, the main object type is looked up here.  If there is a 
# replacement class, the object type is changed to the replacement class.
# If not, then the object is contstructed as is, and a call to type is 
# generated.
# Each entry is a dictionary containing a tuple that lists the new
# type or an empty type and the call to set that option.
typeDefMappings = { \
	"Slider":{"Vert Fill":("Fill_Slider", ""), \
					"Horizontal":("Hor_Slider",""), \
					"Horz Fill":("Hor_Fill_Slider",""), \
					"Horz Knob":("Hor_Nice_Slider",""), \
					"Vert Knob":("Nice_Slider","")},
	"Value_Slider":{"Horizontal":("Hor_Value_Slider",""),
						"Vert Knob":("","type = 4"),
						"Horz Fill":("","type = 3"),
						"Vert Fill":("","type = 2"),
						"Horz Knob":("","type = 5") },
	"Fl_Fill_Slider":{"Horizontal":("Fl_Hor_Fill_Slider","")},
	"Scroll":{"HORIZONTAL":("","type = 1"),
					"VERTICAL":("","type = 2"),
					"BOTH":("","type = 3"),
					"ALWAYS_ON":("","type = 4"),
					"HORIZONTAL_ALWAYS":("","type = 5"),
					"VERTICAL_ALWAYS":("","type = 6"),
					"BOTH_ALWAYS":("","type = 7"),
					},
	"Scrollbar":{"Horizontal":("","type = 1")},
	"Counter":{	"Simple":("Simple_Counter","")},
	"Roller":{	"Horizontal":("","type = 1")},
	"Dial":{		"Line":("Line_Dial",""), 
					"Fill":("Fill_Dial","")},
	"Browser":{  "Select":("Select_Browser",""), \
					"Hold":("Hold_Browser",""), \
					"Multi":("Multi_Browser","")},
	"Button":{	"Radio":("", "type = 102"),
					"Toggle":("", "type = 1")},
        	"Check_Button":{	"Radio":("", "type = 102"),
					"Toggle":("", "type = 1")},
	"Input":{	"Multiline":("Multiline_Input", "")}

	}





class CodePy:
	def __init__(self):
		self.importStatementRegex = re.compile("^(import|from)\s")
		self.handlers = { 'decl':self.doDeclaration, 
					 'class':self.doClass,
					 'func':self.doFunction }


	
		# used doObjectOptions to generate code for object options
		self.binaryOptionSet = { 
			# the format of this dictionary is:  "option":"code to generate"
			# general options:
			"resizable":".resizable = %(name)s",
			"hotspot":":hotspot(%(name)s)",
			"modal":":set_modal()",
			"noborder":".border = 0",
			"deactivate":":deactivate()",
			"whenDef":".when = %(value)s",
			"boxDef":".box = %(value)s",
			# handled in constructor
			# "labelDef":".label='%(value)s'",                        
			"labeltypeDef":".labeltype = %(value)s",
			"labelfontDef":".labelfont = %(value)s",
			"labelsizeDef":".labelsize = %(value)s",
			"labelcolorDef":".labelcolor = %(value)s",
			"alignDef":".align =%(value)s",
			"boxDef":".box = %(value)s",
			"down_boxDef":".down_box = %(value)s",
			"callbackDef":".callback = %(value)s",
			"callbackFuncDef":".callbackFunc = %(value)s",                        
			"shortcutDef":".shortcut = %(value)s",
			"colorDef":".color = %(value)s",
			"selection_colorDef":".selection_color = %(value)s",
			#"user_dataDef":"#user_data(%(value)s) doesn't work, use code",
			# Button specific options:
			"Toggle":".type = 1", 
			"Radio":".type =102",
			# valuator options
			"minimumDef":".minimum = %(value)s",
			"stepDef":".step =%(value)s"

			}
		pass

	def write(self, parseTree, outputFile=stdout):
		# generates python code from 'parseTree' to the
		# opened file 'outputFile'
		self.out = outputFile
		self.tree = parseTree
		self.out.write("-- generated by flconvert\n-- Don't edit by hand.\n\n"
				"-- from fltk import *\n\n")
		if len(parseTree)==0:
			self.write("# (empty data files are somewhat pointless)\n")
			return


		#generated global object defs
		#self.out.write("# global object names\n")
		#for od in self.getGlobalNames():
		#	# (name, type, func)
		#	self.out.write("%s = None      # type '%s' from '%s()'\n"%od)
		self.out.write("\n\n")

		#generate the function defs
		for obj in parseTree:
			self.handlers[obj['def_type']](obj, 0)

		#add Python idiom that allows file to run as a program or
		#to be used as in an import iff there is a 'main' function
		for deff in parseTree:
			if deff['def_type']=='func':
				if deff['name']=="" or deff['name']=="main":
					# self.out.write("function fltk_dispatch(w) if w.callback then w:callback() end end\n")
					self.out.write("window = main(); window:show()\n")
					break
					#self.out.write("\n\n"+
					#	       "if __name__=='__main__':\n"+
					#	       ts(1)+"import sys\n"+
					#	       ts(1)+"window = main()\n"+
					##	       ts(1)+"window.show(len(sys.argv), sys.argv)\n"+
					#	       ts(1)+"Fl.run()\n")
					


	def getGlobalNames(self, tree=None):
		# returns list of global (name, type, functionNameWhereDeclared) tuples 
		# defaults to using the coder's parse tree
		if not tree:
			tree = self.tree
		globalNames = []
		for deff in tree:
			if deff['def_type']=='func':
				for obj in deff['objects']:
					if obj['type'] == "gui":
						globalNames = globalNames + \
							self.getGlobalNamesFromObject(obj, deff['name'])
		return globalNames

	def getGlobalNamesFromObject(self, obj, funcName):
		# returns list of global (name, type, functionNameWhereDeclared) tuples 
		globalNames = []
		if obj.has_key('name') and obj['name'] != '':
				if obj['options'].has_key('classDef'):
					objType = obj['options']['classDef']
				else:
					objType = obj['class']
				globalNames.append( (obj['name'], objType, funcName) )
		if obj.has_key('objects'):
			for subObj in obj['objects']:
				globalNames = globalNames + \
					self.getGlobalNamesFromObject(subObj, funcName)
		return globalNames


	def getImportsFromObject(self, obj, funcName):
		# returns list of 'extra code' lines that are import statements
		# present in the object or its children
		importStatements = []
		if obj['options'].has_key('codeDef'):
			codeDefs = obj['options']['codeDef']
			for lineOfCode in codeDefs:
					if self.importStatementRegex.match(lineOfCode):
						importStatements.append( (lineOfCode, funcName) )
		if obj.has_key('objects'):
			for subObj in obj['objects']:
				importStatements = importStatements + \
					self.getImportsFromObject(subObj, funcName)
		return importStatements


	def getObjectName(self, obj, childLevel, childNumber):
		# returns name used for this object in the generated code
		# it will either be the name defined in fluid or
		# "o_<child level>_<child number> where child level is
		# the level of the objects nesting within the function
		name = obj['name']
		if (name == ''):
			name = "o_%d_%d"%(childLevel, childNumber)
		return name


	def doClass(self, aClass, tabLevel):
		if aClass.has_key('base'):
			self.out.write(ts(tabLevel)+"class %s(%s):\n\n" \
				%(aClass['name'], aClass['base']))
		else:
			self.out.write(ts(tabLevel)+"class %s:\n\n"%(aClass['name']))

		for member in aClass['objects']:
			self.handlers[member['def_type']]( member, tabLevel+1)

		self.out.write("# end of class %s\n\n"%(aClass['name']))
		

	def doDeclaration(self, decl, tabLevel):
		self.out.write("%s%s\n\n"%(ts(tabLevel),decl['code'][1:-1]))
			
	def doFunction(self, func, tabLevel):
		# write Python code for the given function

		# this mimics fluid's behavior
		name = func["name"]
		if name=="":
			name="main"

		# generate the list of input paramters
		inputParams = ""
		for ip in func['inputParams']:
			if (inputParams != ''):
				inputParams = inputParams + ", "
			inputParams = inputParams + ip[0]
		#self.out.write("%sdef %s(%s):\n"%(ts(tabLevel), name, inputParams))
		self.out.write("%sfunction %s(%s)\n"%(ts(tabLevel), name, inputParams))

		#write any global declarations
		#for obj in func['objects']:
		#	for globalVar in self.getGlobalNamesFromObject(obj, ""):
		#		self.out.write(ts(tabLevel+1)+"global "+globalVar[0]+"\n")

		#process each object / line of code
		nameOfReturnedObject = None
		functionsObjects = func['objects']
		for i in range(0, len(functionsObjects), 1):
			obj = functionsObjects[i]
			self.doObject( obj, tabLevel+1, 1, i)
			if (not nameOfReturnedObject) and (obj['type']=='gui'):
				nameOfReturnedObject = self.getObjectName(obj, 1, i)
			

		#generate return statement if need be
		#that is if there were any gui objects created in the function
		if nameOfReturnedObject:
			self.out.write( "\n"+ts(tabLevel+1)+"return "+nameOfReturnedObject)

		self.out.write("\nend\n\n")


	def doObjectOptions(self, opts, tabLevel, name):
		# generates code that handles the different binary object
		# options, such as type, hidden, etc.

		# dictionary used for replacement in the generated code
		# tags in the 'code to generate' portion of binaryOptionSet
		# are replaced by these values
		tags = { "name":name, "value":"TheNode'sAttribute" }	

		# this dictionary is used to fix the 'value' parameter of
		# the class option.  It usually means fixing up the naming
		# between native FLTK constants and their wrapped equivalents.
		# This is where node attributes from the fluid file are
		# mapped to different strings as dictated by the Python wrapper.
		optionValueFix = {
			# node attrite:new text for the attribute
			"NORMAL_LABEL": "Labeltype.normal",
			"NO_LABEL": "Labeltype.none",
			"FREE_LABELTYPE": "Labeltype.free",
			"SYMBOL_LABEL": "Labeltype.symbol",
			"SHADOW_LABEL": "Labeltype.shadow",
			"ENGRAVED_LABEL": "Labeltype.engraved",
			"EMBOSSED_LABEL": "Labeltype.embossed",
			"NO_BOX":"Boxtype.none",
			"FLAT_BOX":"Boxtype.flat",
			"UP_BOX":"Boxtype.up",
			"DOWN_BOX":"Boxtype.down",
			"UP_FRAME":"Frametype.up",
			"DOWN_FRAME":"Frametype.down",
			"THIN_UP_BOX":"Boxtype.thin_up",
			"THIN_DOWN_BOX":"Boxtype.thin_down",
			"THIN_UP_FRAME":"Frametype.thin_up",
			"THIN_DOWN_FRAME":"Frametype.thin_down",
			"ENGRAVED_BOX":"Boxtype.engraved",
			"EMBOSSED_BOX":"Boxtype.embossed",
			"ENGRAVED_FRAME":"Frametype.engraved",
			"EMBOSSED_FRAME":"Frametype.embossed",
			"BORDER_BOX":"Boxtype.border",
			"BORDER_FRAME":"Frametype.border",
			"ROUND_UP_BOX": "Boxtype.round_up",
			"ROUND_DOWN_BOX": "Boxtype.round_down",
			"SHADOW_BOX": "Boxtype.shadow",
			"SHADOW_FRAME": "Frametype.shadow",
			"ROUNDED_BOX": "Boxtype.rounded",
			"ROUNDED_FRAME": "Frametype.rounded",
			"RFLAT_BOX": "Boxtype.rflat",
			"RSHADOW_BOX": "Boxtype.rshadow",
			"DIAMOND_UP_BOX": "Boxtype.diamond_up",
			"DIAMOND_DOWN_BOX": "Boxtype.diamond_down",
			"OVAL_BOX": "Boxtype.oval",
			"OSHADOW_BOX": "Boxtype.oshadow",
			"OVAL_FRAME": "Frametype.oval",
			"OFLAT_BOX": "Boxtype.oflat",
		}

			
		for key in opts.keys():
                        if key=='callbackFuncDef':
                                self.out.write(ts(tabLevel)+"function " + name+":callback() " + opts[key][1:-1] + " end\n")
                                continue
			#if key=='callbackDef':
			#	self.out.write( ts(tabLevel)+name+".callback = " + opts[key] +"\n")
			#	self.out.write(ts(tabLevel)+"fl_cb("+name+")\n")
			#	continue
			if self.binaryOptionSet.has_key(key):
				if  optionValueFix.has_key(opts[key]):
					tags["value"]=optionValueFix[opts[key]]
				else:
					tags["value"]=opts[key]
				self.out.write( ts(tabLevel)+name+"" + 
					self.binaryOptionSet[key]%tags+"\n")

	
	def doObject(self, obj, tabLevel, childLevel=0, childNumber=0, parent=None):
		# write python code for the given object defintion

		# 'code' objects are completely different, so pass of their handling...
		if obj['type']=='code':
			self.doCodeObject(obj, tabLevel, childLevel, childNumber, parent)
			return

		# menus are pretty complicated, so pass off their handling....
		if obj['class'] in fltkBuiltinMenuClasses:
			self.doMenuObject(obj, tabLevel, childLevel, childNumber, parent)	
			return

		# at this point we know obj is a gui object

		self.out.write("\n") # separate each object def w/a line

		opts = obj['options']
		name = self.getObjectName( obj, childLevel, childNumber )


		if opts.has_key('classDef'):
			type = opts['classDef']
		else:
			type = obj['class']
                        if type[0:3] == "Fl_":
                                type = type[3:]
		#handle the special cases of types in the class options section
		#this may change the object's class
		if opts.has_key('typeDef'):
			#self.out.write("\n-- " + typeDefMappings[type] + "\n")
			try:

				newtype = typeDefMappings[type][opts['typeDef']][0]
				if newtype != "":
					type = newtype
			except:
				self.out.write("# warning: following obj has type class\n")
				self.out.write("# option not handled by code generator:\n")	
				self.out.write("# "+str(opts['typeDef'])+"\n")
				self.out.write("# Please email this comment and the following\n")
				self.out.write("# 2 lines to pyfltk@egroups.com\n")

		xywh = opts['xywhDef']  
		constructorParams = "%s, %s, %s, %s"%(xywh[0], xywh[1], 
			xywh[2], xywh[3])
		if opts.has_key('labelDef'):
			constructorParams = '%s, "%s"'%(constructorParams, opts['labelDef'])
		self.out.write(ts(tabLevel))
		if (obj['name'] == ''):
			self.out.write('local ')
		self.out.write( name+" = " + type + \
			"{"+constructorParams+"}\n")
		#handle the special cases of types in the class options section
		#add the 'type' call if needed - becuase not all are subclasses
		if opts.has_key('typeDef'):
			try:
				typeCall = typeDefMappings[type][opts['typeDef']][1]
				if typeCall != "":
					self.out.write( ts(tabLevel)+name+"."+typeCall+"\n")
			except:
				# XXX gideon
				self.out.write("# XXX unknown typedef " + type + ' ' + opts['typeDef'] + '\n')
				pass
					


		#add the children list
		#self.out.write( ts(tabLevel)+name+".pyChildren=[]\n")

		# see NOTES file
		#self.out.write( ts(tabLevel)+name+".thisown = 0\n")

		#recurse and process child objects
		if obj.has_key('objects'):
			#for o in obj['objects']:
			#	self.doObject(o, tabLevel)
			for i in range(0, len(obj['objects']), 1):
				self.doObject( obj['objects'][i], tabLevel, childLevel+1, 
					i, name)

		#generate code for the object's options
		self.doObjectOptions(opts, tabLevel, name)

		
		#generate 'extra code' if it is not some form of import statement
		if opts.has_key("codeDef"):
			self.out.write(ts(tabLevel) + "do -- extra code\n")
			self.out.write(ts(tabLevel+1) + "local self = " + name +"\n")
			codeDefs = opts["codeDef"]
			for lineOfCode in codeDefs:
				if (not self.importStatementRegex.match(lineOfCode)):
					self.out.write(ts(tabLevel+1)+lineOfCode+"\n")
			self.out.write(ts(tabLevel) + "end -- extra code\n")

		#generate the 'end' statement if the object is a group
		if obj['class'] in fltkBuiltinGroupClasses:
			self.out.write( ts(tabLevel)+name+":end_layout()\n" )

		#add to the parent's list of children
		#if (parent):
		#	self.out.write( "%s%s.pyChildren.append(%s)\n"%( ts(tabLevel),
		#			parent, name) )





	def doCodeObject(self, obj, tabLevel, childLevel=0, childNumber=0, 
			parent=None):
		#generates 'code' objects
		#have to split on end line and add any extra indentation
		#splitting blindly on \012 is overly simplistic, but will
		#should handle most cases fine 
		linesOfCode = string.split(obj['code'], "\012")
		for lineOfCode in linesOfCode:
			self.out.write( ts(tabLevel) + lineOfCode + "\n" )

		

	def doMenuObject(self, obj, tabLevel, childLevel=0, childNumber=0, 
			parent=None):
		# write python code for the given menu object defintion
		opts = obj['options']
		name = self.getObjectName( obj, childLevel, childNumber)
		type = obj['class']
		if type[0:3] == "Fl_":
			type = type[3:]

		xywh = opts['xywhDef']
		constructorParams = "%s, %s, %s, %s"%(xywh[0], xywh[1], 
			xywh[2], xywh[3])
		if opts.has_key('labelDef'):
			constructorParams = '%s, "%s"'%(constructorParams, opts['labelDef'])

		self.out.write( ts(tabLevel)+name+" = "+type+"{"+constructorParams+"}\n")
		#add the children 
		# self.out.write( ts(tabLevel)+name+".pyChildren=[]\n")

		# see NOTES file
		# self.out.write( ts(tabLevel)+name+".thisown = 0\n")

		# generate call to set the menu
		if None and obj.has_key('objects') and len(obj['objects'])>0:
			self.out.write("\n"+ts(tabLevel)+"# *hack alert*\n")
			self.out.write(ts(tabLevel)+"setMenu( "+name+", (\n")
			prefix = ts(tabLevel+1)
			for o in obj['objects']:
				self.__dumpMenuSubObject(o, prefix)
			self.out.write(ts(tabLevel+1)+"(None, ) ")
			self.out.write("))\n\n")

		#generate the 'end' statement if the object is a group
		if obj['class'] in fltkBuiltinGroupClasses:
			self.out.write( ts(tabLevel)+name+":end_layout()\n" )

		#add to the parent's list of children
		if (None and parent):
			self.out.write( "%s%s.pyChildren.append(%s)\n"%( ts(tabLevel),
					parent, name) )


	def __dumpMenuSubObject(self, o, prefix):
			# generates one line of menu initializer code
			# for use in a call to setMenu
			opts = o['options']
			label = opts['labelDef']
			if opts.has_key('shortcutDef'):
				accel = opts['shortcutDef'] 
			else:
				accel = '0'

			if opts.has_key('callbackDef'):
				callback = opts['callbackDef']
			else:
				callback = '0'

			#handle menu item modifiers
			#- yes, these hard-coded constants are bad
			flags = 0
			if opts.has_key('deactivate'):
				flags = flags + 1
			if opts.has_key('value'):
				flags = flags + 4
			if opts.has_key('hide'):
				flags = flags + 16
			if o['class']=='submenu':
				flags = flags + 64 
			if opts.has_key('divider'):
				flags = flags + 128

			self.out.write('%s("%s", %s, %s, 0, %d), \n'%( prefix, label,
				accel, callback, flags)	)
		
			if o.has_key('objects'):
				for childObject in o['objects']:
					self.__dumpMenuSubObject(childObject, prefix+ts(1))
				self.out.write(prefix+ts(1)+"(None, ),\n")

	

if __name__ == '__main__':

	from FlScanner import FlScanner
	from FlParser import FlParser
	from FlPreProcessor import FlPreProcessor
	from FlParseTreeConverter import FlParseTreeConverter
	import cStringIO
	from util import *
	from FlParseTreeConverter import dumpAstTree

	try:
		import sys
		filename = sys.argv[1]
	except:
		filename = "test.fl"
	f = open(filename)
	input = f.read()

	tempFile = cStringIO.StringIO()

	preProcessor = FlPreProcessor()
	preProcessor.tokenize(input, tempFile)
	input = tempFile.getvalue()

	print "PreProcesssed:\n"+input
	print
	scanner = FlScanner()
	tokens = scanner.tokenize(input)

	dumpSequence(tokens, "Tokens:")

	print
	print "About to Parse."
	parser = FlParser()
	parseTree =  parser.parse(tokens)

	print "\n\nParse Tree:"
	dumpAstTree(parseTree)

	ptc  = FlParseTreeConverter(parseTree)
	defs = ptc.getDefs()
	#print "\n\nAfter FlParseTreeConverter:"
	#print str(defs)

	#print "\n\nCode:"
	#code = CodePy()
	#code.write(defs, stdout)


